#These are brought into the behaveml namespace
from behaveml.io import read_DLC_tracks, save_DLC_tracks_h5, load_sklearn_model, save_sklearn_model
from behaveml.interpolation import interpolate_lowconf_points
from behaveml.video import VideosetDataFrame, clone_metadata, load_videodataset
from behaveml.features import cnn_probability_feature_maker, mars_feature_maker
from behaveml.plot import plot_tsne
from behaveml.unsupervised import compute_tsne_embedding

#These are the functions imported when doing 'from behaveml import *'
__all__ = [VideosetDataFrame, clone_metadata, read_DLC_tracks, load_videodataset, load_sklearn_model, save_sklearn_model,
           save_DLC_tracks_h5, mars_feature_maker, cnn_probability_feature_maker,
           interpolate_lowconf_points, plot_tsne, compute_tsne_embedding]


import pandas as pd 
import numpy as np

from behaveml.io import XY_IDS

from itertools import product

#Original MABE dataset
#WIDTH = 1024
#HEIGHT = 570

#TODO 
# Allow these to change per video? Would be annoying to implement.
WIDTH = 1600
HEIGHT = 1200

#Shift decorator
#If a feature creation function has this applied then the features are
#automatically shifted/differences/had distribution stats computed and
#concatenated with the rest of the table

#The decorator maker, so we can provide arguments
def augment_features(window_size = 5, n_shifts = 3, mode = 'shift'):
    #The decorator
    def decorator(feature_function):
        #What is called instead of the actual function, assumes the feature making
        #function returns the names of the columns just made
        def wrapper(*args, **kwargs):
            #Compute the features
            #print(args, kwargs, args[0])
            if 'mode' in kwargs:
                mode = kwargs['mode']
            if 'n_shifts' in kwargs:
                n_shifts = kwargs['n_shifts']

            old_cols = set(args[0].columns)
            df = feature_function(*args, **kwargs)
            if n_shifts == 0: return df
            new_cols = set(df.columns)
            added_cols = list(new_cols.difference(old_cols))
            #Shift the features just made
            shifted_data = []
            if mode == 'distr':
                window_sizes = [1, 5, 10]
                for ws in window_sizes:
                    data = np.dstack([np.array(df[added_cols].shift(p)) for p in range(-ws, ws+1)])
                    min_data = pd.DataFrame(np.min(data, axis = 2), columns = [f'{cn}_min_pm_{ws}' for cn in added_cols])
                    max_data = pd.DataFrame(np.max(data, axis = 2), columns = [f'{cn}_max_pm_{ws}' for cn in added_cols])
                    std_data = pd.DataFrame(np.std(data, axis = 2), columns = [f'{cn}_std_pm_{ws}' for cn in added_cols])
                    mean_data = pd.DataFrame(np.mean(data, axis = 2), columns = [f'{cn}_mean_pm_{ws}' for cn in added_cols])
                    shifted_data += [min_data, max_data, std_data, mean_data]
            else:
                periods = [-(i+1)*window_size for i in range(n_shifts)] + \
                        [(i+1)*window_size for i in range(n_shifts)]
                #Rename all column names
                for p in periods:
                    if mode == 'shift':
                        s_df = df[added_cols].shift(p)
                    elif mode == 'diff':
                        s_df = df[added_cols].diff(p)
                    s_df = s_df.rename(columns = {k:f'{k}_shifted_{p}' for k in added_cols})
                    shifted_data.append(s_df)
            #Combine with current table
            #TODO
            # Figure out why a reset_index is needed here... seems to cause issues downstream
            # df has a funny index or column structure?
            #df = pd.concat([df.reset_index(drop = True)] + shifted_data, axis = 1)
            df = pd.concat([df] + shifted_data, axis = 1)
            return df
        return wrapper
    return decorator

from pandas.api.types import is_numeric_dtype

def boiler_plate(features_df):
    reversemap = None
    if 'seq_id' in features_df:
        hashmap = {k: i for (i,k) in enumerate(list(set(features_df['seq_id'])))}
        reversemap = {i: k for (i,k) in enumerate(list(set(features_df['seq_id'])))}
        features_df['seq_id_'] = [hashmap[i] for i in features_df['seq_id']]
        features_df['seq_id'] = features_df['seq_id_']

    to_drop = ['seq_id_', 'Unnamed: 0']
    for col in to_drop:
        if col in features_df.columns:
            features_df = features_df.drop(columns = col)
    #Impute nas
    for col in features_df:
        if is_numeric_dtype(features_df[col]):
            features_df[col] = features_df[col].fillna(features_df[col].mean())
    return features_df, reversemap

def make_features_distances(df, animal_setup):

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']

    feature_set_name = 'features_distances'

    features_df = df.copy()

    ##Make the distance features
    for i, bp1 in enumerate(bodypart_ids):
        for j, bp2 in enumerate(bodypart_ids):
            if i < j:
                for mouse_id in mouse_ids:
                    #We can compute the intra-mouse difference
                    f1x = '_'.join([mouse_id, 'x', bp1])
                    f2x = '_'.join([mouse_id, 'x', bp2])
                    f1y = '_'.join([mouse_id, 'y', bp1])
                    f2y = '_'.join([mouse_id, 'y', bp2])
                    f_new = '_'.join([mouse_id, 'dist', bp1, bp2])
                    features_df[f_new] = \
                        np.sqrt((features_df[f1x] - features_df[f2x])**2 + 
                                (features_df[f1y] - features_df[f2y])**2)
            #Inter-mouse difference
            f1x = '_'.join([mouse_ids[0], 'x', bp1])
            f2x = '_'.join([mouse_ids[1], 'x', bp2])
            f1y = '_'.join([mouse_ids[0], 'y', bp1])
            f2y = '_'.join([mouse_ids[1], 'y', bp2])
            f_new = '_'.join(['M0_M1', 'dist', bp1, bp2])
            features_df[f_new] = \
                        np.sqrt((features_df[f1x] - features_df[f2x])**2 + 
                                (features_df[f1y] - features_df[f2y])**2)

    #Remove base features
    features_df = features_df.drop(columns = colnames)

    ##Clean up seq_id columns
    features_df, reversemap = boiler_plate(features_df)

    return features_df, reversemap, feature_set_name

@augment_features()
def _compute_centroid(df, name, animal_setup, body_parts = None, n_shifts = 3, mode = 'shift'):
    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']

    if body_parts is None:
        body_parts = bodypart_ids

    df = df.copy()
    for mouse_id in mouse_ids:
        part_names_x = [f'{mouse_id}_x_{i}' for i in body_parts]
        part_names_y = [f'{mouse_id}_y_{i}' for i in body_parts]
        df[f'centroid_{name}_{mouse_id}_x'] = np.mean(df[part_names_x], axis = 1)
        df[f'centroid_{name}_{mouse_id}_y'] = np.mean(df[part_names_y], axis = 1)
    return df

@augment_features()
def _compute_abs_angle(df, name, animal_setup, bps, centroid = True, n_shifts = 3, mode = 'shift'):
    mouse_ids = animal_setup['mouse_ids']
    df = df.copy()
    if len(bps) != 2:
        raise ValueError('Abs angle only works between 2 bodyparts, too many or too few specified')
    for mouse_id in mouse_ids:
        if centroid:
            diff_x = df[f'{bps[0]}_{mouse_id}_x'] - df[f'{bps[1]}_{mouse_id}_x']
            diff_y = df[f'{bps[0]}_{mouse_id}_y'] - df[f'{bps[1]}_{mouse_id}_y']
        else:
            diff_x = df[f'{mouse_id}_x_{bps[0]}'] - df[f'{mouse_id}_x_{bps[1]}']
            diff_y = df[f'{mouse_id}_y_{bps[0]}'] - df[f'{mouse_id}_y_{bps[1]}']
        df[f'angle_{name}_{mouse_id}'] = np.arctan2(diff_y,diff_x)  
    return df

@augment_features()
def _compute_rel_angle(df, name, animal_setup, bps, centroid = False, n_shifts = 3, mode = 'shift'):
    mouse_ids = animal_setup['mouse_ids']    
    df = df.copy()
    if len(bps) != 3:
        raise ValueError('too many body parts to compute an absolute angle. Only works for 2')
    for mouse_id in mouse_ids:
        if centroid:
            diff_x1 = df[f'{bps[0]}_{mouse_id}_x'] - df[f'{bps[1]}_{mouse_id}_x']
            diff_y1 = df[f'{bps[0]}_{mouse_id}_y'] - df[f'{bps[1]}_{mouse_id}_y']
            diff_x2 = df[f'{bps[2]}_{mouse_id}_x'] - df[f'{bps[1]}_{mouse_id}_x']
            diff_y2 = df[f'{bps[2]}_{mouse_id}_y'] - df[f'{bps[1]}_{mouse_id}_y']
        else:
            diff_x1 = df[f'{mouse_id}_x_{bps[0]}'] - df[f'{mouse_id}_x_{bps[1]}']
            diff_y1 = df[f'{mouse_id}_y_{bps[0]}'] - df[f'{mouse_id}_y_{bps[1]}']
            diff_x2 = df[f'{mouse_id}_x_{bps[2]}'] - df[f'{mouse_id}_x_{bps[1]}']
            diff_y2 = df[f'{mouse_id}_y_{bps[2]}'] - df[f'{mouse_id}_y_{bps[1]}']

        diff1 = np.vstack((diff_x1, diff_y1)).T
        diff2 = np.vstack((diff_x2, diff_y2)).T
        cosine_angle = np.sum(diff1*diff2, axis = 1) / (np.linalg.norm(diff1, axis = 1) * np.linalg.norm(diff2, axis = 1))
        df[f'angle_{name}_{mouse_id}'] = np.arccos(cosine_angle)
    return df

@augment_features()
def _compute_ellipsoid(df, animal_setup, n_shifts = 3, mode = 'shift'):
    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']

    df = df.copy()
    #Perform SVD
    colnames = ['_'.join([a[0], a[2], a[1]]) for a in product(mouse_ids, bodypart_ids, XY_IDS)]
    data = np.array(df[colnames]).reshape(-1, 2, 7, 2)
    mean_data = np.transpose(np.tile(np.mean(data, axis = 2), (7,1,1,1)), (1,2,0,3))
    svals = np.linalg.svd(data-mean_data, compute_uv = False)
    #Not technically correct, but not sure if the square of the singular values is exaclty
    #what we want either. This keeps the scale roughly the same as the distances involved
    evals = svals
    # evals = (svals*svals)/6
    for idx,m_id in enumerate(mouse_ids):
        df[f'ellipse_major_{m_id}'] = evals[:,idx,0]
        df[f'ellipse_minor_{m_id}'] = evals[:,idx,1]
        ## ratio of major and minor
        df[f'ellipse_ratio_{m_id}'] = df[f'ellipse_minor_{m_id}']/df[f'ellipse_major_{m_id}']
        ## area of ellipse
        df[f'ellipse_area_{m_id}'] = df[f'ellipse_minor_{m_id}']*df[f'ellipse_major_{m_id}']

    ## ratio of areas of ellipses of the mice
    df[f'ellipse_area_ratio'] = df[f'ellipse_area_{mouse_ids[0]}']/df[f'ellipse_area_{mouse_ids[1]}']

    return df

#Recall framerate is 30 fps
def _compute_kinematics(df, names, animal_setup, window_size = 5, n_shifts = 3):
    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']
    df = df.copy()
    for mouse_id in mouse_ids:
        for name in names:
            ## Speed of centroids
            dx = df[f'centroid_{name}_{mouse_id}_x'].diff(window_size)
            dy = df[f'centroid_{name}_{mouse_id}_y'].diff(window_size)
            df[f'centroid_{name}_{mouse_id}_speed'] = np.sqrt(dx**2 + dy**2)
            colnames.append(f'centroid_{name}_{mouse_id}_speed')
            ## Acceleration of centroids
            ddx = dx.diff(window_size)
            ddy = dy.diff(window_size)
            df[f'centroid_{name}_{mouse_id}_accel_x'] = ddx/(window_size**2)
            df[f'centroid_{name}_{mouse_id}_accel_y'] = ddy/(window_size**2)
    return df

@augment_features()
def _compute_relative_body_motions(df, animal_setup, window_size = 3, n_shifts = 3, mode = 'shift'):

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']
    
    #Compute vector connecting two centroids
    dx = df[f'centroid_all_{mouse_ids[0]}_x'] - df[f'centroid_all_{mouse_ids[1]}_x']
    dy = df[f'centroid_all_{mouse_ids[0]}_y'] - df[f'centroid_all_{mouse_ids[1]}_y']
    dm = np.sqrt(dx**2 + dy**2)
    df['distance_main_centroid'] = dm

    #Compute velocity of mouse centroids
    for m_id in mouse_ids:
        vx = df[f'centroid_all_{m_id}_x'].diff(window_size)/window_size
        vy = df[f'centroid_all_{m_id}_y'].diff(window_size)/window_size
        v_tangent = (dx*vx + dy*vy)/dm
        v_perp_x = vx - dx*v_tangent/dm
        v_perp_y = vy - dy*v_tangent/dm
        v_perp = np.sqrt(v_perp_x**2 + v_perp_y**2)
        df[f'relative_vel_tanget_{m_id}'] = v_tangent
        df[f'relative_vel_perp_{m_id}'] = v_perp

        ## relative distance scaled
        #Distance between main centroids, divided by length of major axis of each mouse
        df[f'scaled_main_centroid_distance_by_ellipse_major_{m_id}'] = dm/df[f'ellipse_major_{m_id}']

    return df

@augment_features()
def _compute_relative_body_angles(df, animal_setup, n_shifts = 3, mode = 'shift'):

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']

    for idx, m_id in enumerate(mouse_ids):
        #Compute vector connecting two centroids
        dx1 = df[f'centroid_all_{mouse_ids[1-idx]}_x'] - df[f'centroid_all_{mouse_ids[idx]}_x']
        dy1 = df[f'centroid_all_{mouse_ids[1-idx]}_y'] - df[f'centroid_all_{mouse_ids[idx]}_y']

        #Relative angle between body of mouse and line connecting two centroids
        dx2 = df[f'centroid_head_{m_id}_x'] - df[f'centroid_body_{m_id}_x']
        dy2 = df[f'centroid_head_{m_id}_y'] - df[f'centroid_body_{m_id}_y']            
        diff1 = np.vstack((dx1, dy1)).T
        diff2 = np.vstack((dx2, dy2)).T
        cosine_angle = np.sum(diff1*diff2, axis = 1) / (np.linalg.norm(diff1, axis = 1) * np.linalg.norm(diff2, axis = 1))
        df[f'angle_head_body_centroid_{m_id}'] = np.arccos(cosine_angle)

        #Angle between head orientation of one mouse and line connecting two centroids
        dx2 = df[f'{m_id}_x_nose'] - df[f'{m_id}_x_neck']
        dy2 = df[f'{m_id}_x_nose'] - df[f'{m_id}_y_neck']            
        diff1 = np.vstack((dx1, dy1)).T
        diff2 = np.vstack((dx2, dy2)).T
        cosine_angle = np.sum(diff1*diff2, axis = 1) / (np.linalg.norm(diff1, axis = 1) * np.linalg.norm(diff2, axis = 1))
        df[f'angle_head_centroid_{m_id}'] = np.arccos(cosine_angle)

        #Just threshold on if the angle is less than pi/4 radians
        df[f'{mouse_ids[1-idx]}_in_view_of_{mouse_ids[idx]}'] = (cosine_angle > 1/np.sqrt(2)).astype(float)

    return df
    
@augment_features()
def _compute_iou(df, animal_setup, n_shifts = 3, mode = 'shift'):

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']

    mins = {}
    maxs = {}

    for m_id in mouse_ids:
        for xy in XY_IDS:
            colnames = ['_'.join([m_id, xy, bp]) for bp in bodypart_ids]
            mins['_'.join([m_id, xy])] = np.min(df[colnames], axis = 1)
            maxs['_'.join([m_id, xy])] = np.max(df[colnames], axis = 1)

    dx = np.minimum(maxs[f'{mouse_ids[0]}_x'], maxs[f'{mouse_ids[1]}_x']) - np.maximum(mins[f'{mouse_ids[0]}_x'], mins[f'{mouse_ids[1]}_x'])
    dy = np.minimum(maxs[f'{mouse_ids[0]}_y'], maxs[f'{mouse_ids[1]}_y']) - np.maximum(mins[f'{mouse_ids[0]}_y'], mins[f'{mouse_ids[1]}_y'])
    dx = np.maximum(0, dx)
    dy = np.maximum(0, dy)

    bb1_area = (maxs[f'{mouse_ids[0]}_x'] - mins[f'{mouse_ids[0]}_x'])*(maxs[f'{mouse_ids[0]}_y'] - mins[f'{mouse_ids[0]}_y'])
    bb2_area = (maxs[f'{mouse_ids[1]}_x'] - mins[f'{mouse_ids[1]}_x'])*(maxs[f'{mouse_ids[1]}_y'] - mins[f'{mouse_ids[1]}_y'])
    intersection_area = dx*dy
    iou = intersection_area / (bb1_area + bb2_area - intersection_area)
    df['iou'] = iou
    return df

##Distance from centroid of the mouse to the closest vertical edge
## and closest horizontal edge
##Distance to the closest edge
#These depend on the video you're applying it to...
#Which can change from video to video, train to test, etc. So perhaps not be 
@augment_features()
def _compute_cage_distances(features_df, animal_setup, n_shifts = 3, mode = 'shift'):
    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']
    colnames = animal_setup['colnames']
    
    for m_id in mouse_ids:
        features_df[f'centroid_all_{m_id}_x_inverted'] = WIDTH - features_df[f'centroid_all_{m_id}_x']
        features_df[f'centroid_all_{m_id}_y_inverted'] = HEIGHT - features_df[f'centroid_all_{m_id}_y']
        features_df[f'{m_id}_closest_x'] = np.min(features_df[[f'centroid_all_{m_id}_x_inverted', f'centroid_all_{m_id}_x']], axis = 1)
        features_df[f'{m_id}_closest_y'] = np.min(features_df[[f'centroid_all_{m_id}_y_inverted', f'centroid_all_{m_id}_y']], axis = 1)
        features_df[f'{m_id}_closest'] = np.min(features_df[[f'{m_id}_closest_x', f'{m_id}_closest_y']], axis = 1)
        features_df = features_df.drop(columns = [f'centroid_all_{m_id}_x_inverted', f'centroid_all_{m_id}_y_inverted'])
    return features_df

def make_features_mars(df, animal_setup, n_shifts = 3, mode = 'shift', feature_set_name = 'features_mars'):

    features_df = df.copy()

    #######################
    ## Position features ##
    #######################
    features_df = _compute_centroid(features_df, 'all', animal_setup, n_shifts = n_shifts, mode = mode)
    features_df = _compute_centroid(features_df, 'head', animal_setup, ['nose', 'leftear', 'rightear', 'neck'], n_shifts = n_shifts, mode = mode)
    features_df = _compute_centroid(features_df, 'hips', animal_setup, ['lefthip', 'tail', 'righthip'], n_shifts = n_shifts, mode = mode)
    features_df = _compute_centroid(features_df, 'body', animal_setup, ['neck', 'lefthip', 'righthip', 'tail'], n_shifts = n_shifts, mode = mode)

    features_df = _compute_cage_distances(features_df, animal_setup, n_shifts = n_shifts, mode = mode)

    #####################
    #Appearance features#
    #####################

    ## absolute orientation of mice
    features_df = _compute_abs_angle(features_df, 'head_hips', animal_setup, ['centroid_head', 'centroid_hips'], n_shifts = n_shifts, mode = mode)
    features_df = _compute_abs_angle(features_df, 'head_nose', animal_setup, ['neck', 'nose'], centroid = False, n_shifts = n_shifts, mode = mode)
    features_df = _compute_abs_angle(features_df, 'tail_neck', animal_setup, ['tail', 'neck'], centroid = False, n_shifts = n_shifts, mode = mode)
    ## relative orientation of mice
    features_df = _compute_rel_angle(features_df, 'leftear_neck_rightear', animal_setup, ['leftear', 'neck', 'rightear'], n_shifts = n_shifts, mode = mode)
    ## major axis len, minor axis len of ellipse fit to mouses body
    features_df = _compute_ellipsoid(features_df, animal_setup, n_shifts = n_shifts, mode = mode)

    #####################
    #Locomotion features#
    #####################

    features_df = _compute_kinematics(features_df, ['all', 'head', 'hips', 'body'], animal_setup)

    #################
    #Social features#
    #################

    features_df = _compute_relative_body_motions(features_df, animal_setup, n_shifts = n_shifts, mode = mode)
    features_df = _compute_relative_body_angles(features_df, animal_setup, n_shifts = n_shifts, mode = mode)

    #Intersection of union of bounding boxes of two mice
    features_df = _compute_iou(features_df, animal_setup, n_shifts = n_shifts, mode = mode)

    ## distance between all pairs of keypoints of each mouse
    features_df, reversemap, _ = make_features_distances(features_df, animal_setup)

    return features_df, reversemap, feature_set_name

make_features_mars_distr = lambda x, y: make_features_mars(x, y, n_shifts = 3, mode = 'distr', feature_set_name = 'features_mars_distr')

def make_features_mars_w_1dcnn_features(df, animal_setup, n_shifts = 3, mode = 'shift', 
                                         feature_set_name = 'features_mars_distr_w_1dcnn',
                                         fn_in = './data/intermediate/deep_learning_stacking_prediction_probabilities_baseline_test_run_distances.npy'):

    features_df, reversemap, _ = make_features_mars_distr(df, animal_setup)
    forwardmap = {i:k for k,i in reversemap.items()}

    #Probabilities
    all_pred_probs = np.load(fn_in, allow_pickle = True).item()
    #Rewrite these keys
    keys = list(all_pred_probs.keys())
    for k in keys:
        all_pred_probs[forwardmap[k]] = all_pred_probs[k]
        del all_pred_probs[k]

    new_cols = ['1dcnn_prob_0', '1dcnn_prob_1', '1dcnn_prob_2', '1dcnn_prob_3']

    features_df[new_cols] = np.nan

    for k,v in all_pred_probs.items():
        features_df.loc[features_df['seq_id'] == k,new_cols] = v

    return features_df, reversemap, feature_set_name

make_features_mars_w_1dcnn_features_test = lambda x,y: make_features_mars_w_1dcnn_features(x, y, n_shifts = 3, mode = 'shift', feature_set_name = 'features_mars_distr_w_1dcnn',
                                         fn_in = './data/intermediate/deep_learning_stacking_prediction_probabilities_test_baseline_test_run_distances.npy')


from .dl_generators import *

feature_spaces = {'distances': [features_distances, [91]], 
                    'identity': [features_identity, [2,7,2]],
                    'distances_normalized': [features_distances_normalized, [91]],
                    'mars': [features_mars, []],
                    'mars_distr': [features_mars_distr, []],
                }
                    # This one was removed....I think it's not needed
                    # 'mars_no_shift': [features_mars_no_shift, [160]]

sweeps_baseline = {
    'test_run_distances':
        [{'num_samples': 1},
        {
            "model_param__learning_rate": 0.0001,
            "epochs": 15,
            "features": "distances",
            "model_param__dropout_rate": 0.5
        }]
}


from sklearn.preprocessing import StandardScaler
from sklearn.manifold import TSNE
import numpy as np

def compute_tsne_embedding(data, cols, N_rows = 20000):
    tsne_data = StandardScaler().fit_transform(data[cols])
    random_indices = np.random.choice(tsne_data.shape[0], N_rows, replace = False)
    tsne_data = tsne_data[random_indices, :]
    tsne_embedding = TSNE(n_components=2, init = 'pca').fit_transform(tsne_data)
    return tsne_embedding, random_indices


""" Loading and saving tracking and behavior annotation files """
import pandas as pd
import pickle
import numpy as np
from itertools import product
from joblib import dump, load
import os
from glob import glob

XY_IDS = ['x', 'y']
XYLIKELIHOOD_IDS = ['x', 'y', 'likelihood']

def uniquifier(seq):
    """Return a sequence (e.g. list) with unique elements only, but maintaining original list order"""
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def _list_replace(ls, renamer):
    """Replace elements in a list according to provided dictionary"""
    for i, word in enumerate(ls):
        if word in renamer.keys():
            ls[i] = renamer[word]
    return ls

def save_sklearn_model(model, fn_out):
    dump(model, fn_out) 

def load_sklearn_model(fn_in):
    model = load(fn_in)
    return model 

def read_DLC_tracks(fn_in : str, 
                    part_renamer : dict = None, 
                    animal_renamer : dict = None,
                    read_likelihoods : bool = True) -> tuple:
    """Read in tracks from DLC.

    Args:
        fn_in: csv file that has DLC tracks
        part_renamer: dictionary to rename body parts, if needed 
        animal_renamer: dictionary to rename animals, if needed
        read_likelihoods: default True. Whether to attach DLC likelihoods to table

    Returns:
        Pandas DataFrame with (n_animals*2*n_body_parts) columns plus with filename and frame, 
            List of body parts,
            List of animals,
            Columns names for DLC tracks (excluding likelihoods, if read in),
            Scorer
    """
    df = pd.read_csv(fn_in, header = [0,1,2,3], index_col = 0)
    df.columns = df.columns.set_names(['scorer', 'individuals', 'bodyparts', 'coords'])

    scorer = df.columns.get_level_values(0)[0]

    cols = list(df.columns)
    animals = uniquifier([i[1] for i in cols])
    body_parts = uniquifier([i[2] for i in cols])

    n_body_parts = len(body_parts)
    n_animals = len(animals)

    dlc_tracks = np.array(df)
    n_rows = dlc_tracks.shape[0]
    selected_cols = [[3*i, 3*i+1] for i in range(n_body_parts*n_animals)]
    #This line flattens the array here... may be clearer using numpy reshape?
    selected_cols = [j for i in selected_cols for j in i]

    prob_cols = [3*i+2 for i in range(n_body_parts*n_animals)]
    dlc_probs = dlc_tracks[:,prob_cols]
    dlc_tracks = dlc_tracks[:,selected_cols]

    #Put in shape:
    # (frame, animal, x/y coord, body part)
    dlc_tracks = dlc_tracks.reshape((n_rows, n_animals, n_body_parts, 2))
    dlc_tracks = dlc_tracks.transpose([0, 1, 3, 2])

    dlc_probs = dlc_probs.reshape((n_rows, n_animals, n_body_parts, 1))
    dlc_probs = dlc_probs.transpose([0, 3, 1, 2])

    #If we're going to rename items in the list, do it here
    if part_renamer:
        body_parts = _list_replace(body_parts, part_renamer)

    if animal_renamer:
        animals = _list_replace(animals, animal_renamer)

    colnames = ['_'.join(a) for a in product(animals, XY_IDS, body_parts)]
    prob_colnames = ['_'.join(a) for a in product(['likelihood'], animals, body_parts)]

    dlc_probs = dlc_probs.reshape((n_rows, -1))
    final_probs = pd.DataFrame(dlc_probs, columns = prob_colnames)

    dlc_tracks = dlc_tracks.reshape((n_rows, -1))
    final_df = pd.DataFrame(dlc_tracks, columns = colnames)

    if read_likelihoods:
        final_df = pd.concat([final_df, final_probs], axis = 1)

    final_df['filename'] = fn_in
    final_df['frame'] = final_df.index.copy()

    return final_df, body_parts, animals, colnames, scorer

def rename_df_cols(df : pd.DataFrame, renamer : dict) -> pd.DataFrame:
    """Rename dataframe columns 
    
    Args:
        df: Pandas dataframe whose columns to rename
        renamer: dictionary whose key:value pairs define the substitutions to make

    Returns:
        The dataframe with renamed columns.
    """
    return df.rename(columns = renamer)

def save_DLC_tracks_h5(df : pd.DataFrame, fn_out : str) -> None:
    """Save DLC tracks in h5 format.
    
    Args:
        df: Pandas dataframe to save
        fn_out: Where to save the dataframe
    """
    df.to_hdf(fn_out, "df_with_missing", format = 'table', mode="w")

def load_data(fn : str):
    """Load an object from a pickle file
    
    Args:
        fn: The filename

    Returns:
        The pickled object.
    """
    with open(fn, 'rb') as handle:
        object = pickle.load(handle)
    return object 

def _make_sample_dataframe(fn_out = 'sample_dataframe.pkl'):
    from behaveml import VideosetDataFrame, clone_metadata

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    tracking_files = sorted(glob(cur_dir + '/data/dlc/*.csv'))
    boris_files = sorted(glob(cur_dir + '/data/boris/*.csv'))
    frame_length = None              # (float) length of entire horizontal shot
    units = None                     # (str) units frame_length is given in
    fps = 30                         # (int) frames per second
    resolution = (1200, 1600)        # (tuple) HxW in pixels
    metadata = clone_metadata(tracking_files, 
                          label_files = boris_files, 
                          frame_length = frame_length, 
                          fps = fps, 
                          units = units, 
                          resolution = resolution)

    dataset = VideosetDataFrame(metadata)
    path_out = os.path.join(cur_dir, 'data', fn_out)
    to_save = {'dataset': dataset, 'metadata': metadata}
    with open(path_out, 'wb') as handle:
        pickle.dump(to_save, handle, protocol=pickle.HIGHEST_PROTOCOL)

def get_sample_data():
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    path_in = os.path.join(cur_dir, 'data', 'sample_dataframe.pkl')
    with open(path_in, 'rb') as handle:
        b = pickle.load(handle)
    return b['dataset']

def read_boris_annotation(fn_in : str, fps : int, duration : float) -> np.ndarray:
    """Read behavior annotation from BORIS exported csv file
    
    Args:
        fn_in: The filename with BORIS behavior annotations to load
        fps: Frames per second of video
        duration: Duration of video
    
    Returns:
        A numpy array which indicates, for all frames, if behavior is occuring (1) or not (0)
    """
    n_bins = int(duration*fps)
    boris_labels = pd.read_csv(fn_in, skiprows = 15)
    boris_labels['index'] = (boris_labels.index//2)
    boris_labels = boris_labels.pivot_table(index = 'index', columns = 'Status', values = 'Time').reset_index()
    boris_labels = list(np.array(boris_labels[['START', 'STOP']]))
    boris_labels = [list(i) for i in boris_labels]
    ground_truth = np.zeros(n_bins)
    for start, end in boris_labels:
        ground_truth[int(start*fps):int(end*fps)] = 1
    return ground_truth


import matplotlib.pyplot as plt

def plot_tsne(tsne_embedding, color_data, size = 1, figsize = (8,8), legend_title = 'prediction'):
    """asdf
    
    """

    fig, axes = plt.subplots(1,1, figsize = figsize)
    sctr = axes.scatter(x = tsne_embedding[:,0], y = tsne_embedding[:,1], c = color_data, s = size)
    axes.set_xlabel('TSNE dim 1')
    axes.set_ylabel('TSNE dim 2')
    legend1 = axes.legend(*sctr.legend_elements(), loc="lower left", title=legend_title)
    axes.add_artist(legend1)
    return fig, axes


import pandas as pd 
import numpy as np
import pickle 

from joblib import load

import sklearn.metrics 
from sklearn.model_selection import GroupKFold
from sklearn.metrics import f1_score
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.pipeline import Pipeline 
from sklearn.preprocessing import StandardScaler

from behaveml.dl.feature_engineering import make_features_mars_w_1dcnn_features, \
                                            make_features_mars_w_1dcnn_features_test, \
                                            make_features_mars_distr

def compute_mars_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df, _, _ = make_features_mars_distr(df[raw_col_names], animal_setup) 
    return features_df

def make_stacked_features(train_df, test_df): # pragma: no cover

    mars_features_df, reversemap, _ = make_features_mars_w_1dcnn_features(train_df)
    mars_features_df_test, reversemap_test, _ = make_features_mars_w_1dcnn_features_test(test_df)

    features_name = 'features_mars_distr_stacked_w_1dcnn'
    
    n_folds = 5

    X = mars_features_df.drop(columns = ['annotation', 'seq_id'])
    y = mars_features_df['annotation']
    groups = mars_features_df['seq_id']

    X_test = mars_features_df_test.drop(columns = 'seq_id')
    groups_test = mars_features_df_test['seq_id']

    # Setup the CV folds
    cv_groups = GroupKFold(n_splits = n_folds)
    scorer = sklearn.metrics.make_scorer(f1_score, labels = [0,1,2], average = 'macro')

    all_base_models = []

    validation_groups_by_fold = []

    for fold_idx, (train_index, val_index) in enumerate(cv_groups.split(X, y, groups)):

        #fold_idx = 0; train_index, val_index = next(cv_groups.split(X, y, groups))

        X_train = np.array(X)[train_index]
        y_train = np.array(y)[train_index]
        groups_train = groups[train_index]

        y_val = np.array(y)[val_index]
        X_val = np.array(X)[val_index]
        groups_val = groups[val_index]

        validation_groups_by_fold.append(groups_val)

        # Define a bunch of models for the stacking, one for each fold
        level0 = list()
        level0.append(['bayes', Pipeline([('scaler', StandardScaler()), ('nb', GaussianNB())]), [], []])
        level0.append(['et_entropy', ExtraTreesClassifier(n_estimators = 100, criterion='entropy', n_jobs = 5, verbose = 1, random_state = 42), [], []])
        level0.append(['et_gini', ExtraTreesClassifier(n_estimators = 100, criterion='gini', n_jobs = 5, verbose = 1, random_state = 42), [], []])
        level0.append(['rf_entropy', RandomForestClassifier(n_estimators = 20, criterion='entropy', n_jobs = 5, random_state = 42), [], []])
        level0.append(['rf_gini', RandomForestClassifier(n_estimators = 20, criterion='gini', n_jobs = 5, random_state = 42), [], []])

        for idx in range(len(level0)):
            name, model, _, _ = level0[idx]
            print(f"Fitting {name} in fold {fold_idx}")
            #Train the models on each fold, make prediction on held-out for each fold
#            model.fit(X_train, y_train)

            #Save the models too
            model = load(f'./results/level_0_model_{idx}_fold_{fold_idx}.joblib')

            predict_proba_val = model.predict_proba(X_val)
            predict_proba_test = model.predict_proba(X_test)
            #Save this
            level0[idx][2] = predict_proba_val
            level0[idx][3] = predict_proba_test

        all_base_models.append(level0)

    # Once done, concat predictions to make train data for XGB. 
    # And run the inference with the mean of each model to make the test data

    all_val_preds = []
    all_test_preds = {}
    for fold_idx in range(n_folds):
        n_row = validation_groups_by_fold[fold_idx].shape[0]
        base_models = all_base_models[fold_idx]
        col_names = []
        val_preds = np.zeros((n_row, 4*len(level0)))
        for model_idx, model in enumerate(base_models):


            name = model[0]
            classifier = model[1]
            this_val_pred = model[2]
            this_test_pred = model[3]

            if name in all_test_preds:
                all_test_preds[name] += this_test_pred
            else:
                all_test_preds[name] = this_test_pred

            col_names += [f'{name}_{i}' for i in range(4)]
            val_preds[:,(model_idx*4):((model_idx+1)*4)] = this_val_pred

        val_preds = pd.DataFrame(val_preds, columns = col_names)
        val_preds['seq_id'] = validation_groups_by_fold[fold_idx].reset_index()['seq_id']
        all_val_preds.append(val_preds)

    for name in all_test_preds:
        all_test_preds[name] /= n_folds

    all_test_data = np.hstack([v for v in all_test_preds.values()])
    all_test_df = pd.DataFrame(all_test_data, columns = col_names)
    all_test_df['seq_id'] = groups_test

    #Concat all the validation predictions, along with the group names for that fold
    all_val_preds = pd.concat(all_val_preds)

    # Append this to the mars_distr_w_1dcnn file so we have everything together to train the XGB model on
    mars_features_df[col_names] = np.nan
    seq_ids = pd.unique(mars_features_df['seq_id'])
    for seq_id in seq_ids:
        mars_features_df.loc[mars_features_df['seq_id'] == seq_id, col_names] = np.array(all_val_preds.loc[all_val_preds['seq_id'] == seq_id, col_names])

    # Append this to the mars_distr_w_1dcnn test files so we have everything together to test the model
    mars_features_df_test[col_names] = np.nan
    seq_ids = pd.unique(mars_features_df_test['seq_id'])
    for seq_id in seq_ids:
        mars_features_df_test.loc[mars_features_df_test['seq_id'] == seq_id, col_names] = np.array(all_test_df.loc[all_test_df['seq_id'] == seq_id, col_names])

    # Save the feature files
    mars_features_df.to_csv(f'./data/intermediate/train_{features_name}.csv', index = False)
    with open(f'./data/intermediate/test_map_{features_name}.pkl', 'wb') as handle:
        pickle.dump(reversemap_test, handle, protocol=pickle.HIGHEST_PROTOCOL)

    print('Saving test feature file (may take a while)')
    mars_features_df_test.to_csv(f'./data/intermediate/test_{features_name}.csv', index = False)

    return mars_features_df, reversemap, name, mars_features_df_test, reversemap_test

def compute_mars_features_(df : pd.DataFrame, raw_col_names : list, animal_setup : dict): # pragma: no cover

    train_df = pd.read_csv('./data/intermediate/train_df.csv')
    test_df = pd.read_csv('./data/intermediate/test_df.csv')

    #Make features
    train_features, _, name, test_features, test_map = make_stacked_features(train_df, test_df)

    return pd.DataFrame(df) 



import warnings

try:
    from keras.models import Sequential
    import keras.layers as layers
    import keras
    has_keras = True
except ImportError:
    warnings.warn("Keras not found. Deep learning-based features are not available", RuntimeWarning)
    has_keras = False

def build_baseline_model(input_dim, layer_channels=(512, 256), dropout_rate=0., 
                        learning_rate=1e-3, conv_size=5, num_classes = 4,
                        class_weight = None):

    if not has_keras:
        raise RuntimeError("Keras not found. Deep learning-based features are not available")

    print("Building baseline 1D CNN model with parameters:")
    print(f"dropout_rate: {dropout_rate}, learning_rate: {learning_rate}, layer_channels: {layer_channels}, conv_size: {conv_size}")

    def add_dense_bn_activate(model, out_dim, activation='relu', drop=0.):
        model.add(layers.Dense(out_dim))
        model.add(layers.BatchNormalization())
        model.add(layers.Activation('relu'))
        if drop > 0:
            model.add(layers.Dropout(rate=drop))
        return model

    def add_conv_bn_activate(model, out_dim, activation='relu', conv_size=3, drop=0.):
        model.add(layers.Conv1D(out_dim, conv_size))
        model.add(layers.BatchNormalization())
        model.add(layers.Activation('relu'))
        model.add(layers.MaxPooling1D(2, 2))
        if drop > 0:
            model.add(layers.Dropout(rate=drop))
        return model

    model = Sequential()
    model.add(layers.Input(input_dim))
    model.add(layers.BatchNormalization())
    for ch in layer_channels:
        model = add_conv_bn_activate(model, ch, conv_size=conv_size,
                                     drop=dropout_rate)
    model.add(layers.Flatten())
    model.add(layers.Dense(num_classes, activation='softmax'))

    #metrics = [F1Score(num_classes=num_classes, average = 'macro', labels = [0,1,2]), 'accuracy']
    metrics = []
    optimizer = keras.optimizers.Adam(lr=learning_rate)
    model.compile(loss='categorical_crossentropy', 
                optimizer=optimizer,
                metrics=metrics)

    return model 


""" Basic video tracking and behavior class that houses data """

import numpy as np
from sklearn.base import ClassifierMixin, TransformerMixin
from sklearn.metrics import f1_score

def _logit(p):
    return np.log(p / (1 - p))

def _sample_prob_simplex(n=4):
    x = sorted(np.append(np.random.uniform(size = n-1), [0,1]))
    y = np.diff(np.array(x))
    return y

try:
    import ssm 
    class HMMSklearn(ssm.HMM):

        def __init__(self, D, C = 11):
            """HMM model from Linderman state-space model package ssm, tweaked slightly to fit with sklearn syntax
            
            Args:
                D: number of behavioral categories
                C: number of bins to discretize
            """

            self.D = D 
            self.C = C
            super().__init__(D, D + 1, observations = 'categorical', observation_kwargs = {'C': C})

        def fit(self, X, y):

            preds = np.argmax(X, axis = -1)
            X = np.hstack(((X*(self.C-1)).astype(int), np.atleast_2d((preds).astype(int)).T))

            N = len(y)
            transition_matrix = np.ones((self.D, self.D))

            for idx in range(N):
                if idx == 0: continue
                transition_matrix[y[idx-1], y[idx]] += 1
            
            for j in range(self.D):
                transition_matrix[j] /= np.sum(transition_matrix[j]) 

            self.transitions.params = [np.log(transition_matrix)]

            emission_dist = np.ones((self.D, self.D+1, self.C))
            for i in range(self.D):
                for j in range(self.D+1):
                    for k in range(self.C):
                        ct = np.sum(X[(y == i),j] == k)
                        emission_dist[i, j, k] = max(1, ct)
                    emission_dist[i,j,:] /= np.sum(emission_dist[i,j,:])

            self.observations.params = _logit(emission_dist)

        def predict(self, X):
            preds = np.argmax(X, axis = -1)
            X = np.hstack(((X*(self.C-1)).astype(int), np.atleast_2d((preds).astype(int)).T))
            return self.most_likely_states(X)

except ImportError:
    print("Couldn't find ssm module. HMMSklearn model not available. See install instructions: https://github.com/lindermanlab/ssm")

class F1Optimizer(ClassifierMixin):
    def __init__(self, N = 1000, labels = [1]):
        self.N = N
        self.labels = labels

    def fit(self, X, y): #train_labels, train_pred_prob):

        self.dim_x = X.shape[1]

        f = lambda w: f1_score(y, np.argmax((X*w), axis = -1))#, average = 'macro', labels = self.labels)

        w_star = np.ones(self.dim_x)/self.dim_x
        f_star = 0

        for _ in range(self.N):
            w = _sample_prob_simplex(n = self.dim_x)
            f_curr = f(w)
            if f_curr > f_star:
                w_star = w
                f_star = f_curr

        self.w_star = w_star
        self.f_star = f_star

    def predict(self, X):
        return np.argmax(X*self.w_star, axis = -1)

    def predict_proba(self, X):
        return X*self.w_star

    def transform(self, X):
        return self.predict_proba(X)

    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X)

class ModelTransformer(ClassifierMixin):
    def __init__(self, Model, *args, **kwargs):
        self.model = Model(*args, **kwargs)    
        
    def fit(self, X, y):
        self.model.fit(X, y)    
        
    def transform(self, X):
        return self.model.predict_proba(X)
                
    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)


""" Basic video tracking and behavior class that houses data """

import pandas as pd
import pickle 
import os
import numpy as np
import re
from glob import glob 
from sklearn.model_selection import PredefinedSplit

from behaveml.features import Features

from behaveml.io import read_DLC_tracks, XY_IDS, read_boris_annotation, uniquifier
from behaveml.utils import checkFFMPEG

class MLDataFrame(object): # pragma: no cover
    """
    DataFrame useful for interfacing between pandas and sklearn. Stores a data
    table and metadata dictionary. When feature columns, label columns and fold columns are specified
    then creates properties features, labels, folds and splitter that sklearn accepts for ML.
    """
    def __init__(self, data : pd.DataFrame,
                       metadata : dict = {},
                       fold_cols = None,
                       feature_cols = None,
                       label_cols = None):
        self.data = data 
        self.metadata = metadata
        self.fold_cols = fold_cols
        self.feature_cols = feature_cols
        self.label_cols = label_cols

    def add_data(self, new_data, col_names):
        if type(new_data) == pd.DataFrame:
            self.data[col_names] = new_data
        else:
            self.data[col_names] = pd.DataFrame(new_data, index=self.data.index)

    @property
    def features(self):
        if self.feature_cols is None:
            return None
        else:
            return self.data[self.feature_cols].to_numpy()

    @property
    def labels(self):
        if self.label_cols is None:
            return None
        else:
            return self.data[self.label_cols].to_numpy()

    @property
    def folds(self):
        if self.fold_cols is None:
            return None
        else:
            return self.data[self.fold_cols].to_numpy()

    @property
    def splitter(self):
        if self.fold_cols is None:
            return None
        else:
            return self._make_predefined_split(self.folds)

    def _make_predefined_split(self, folds):
        test_indices = np.sum(folds, axis=1) == 0
        test_fold = -1*(test_indices)
        test_fold[test_indices == False] = np.argmin(folds[test_indices == False], axis = 1)
        return PredefinedSplit(test_fold)

    def save(self, fn):
        with open(fn, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def __repr__(self):
        return str(self.data)

def _add_item_to_dict(tracking_files, metadata, k, item):
    for fn in tracking_files:
        metadata[fn][k] = item

def _add_items_to_dict(tracking_files, metadata, k, items):
    for fn, item in zip(tracking_files, items):
        metadata[fn][k] = item

def clone_metadata(tracking_files : list, **kwargs) -> dict:
    """
    Prepare a metadata dictionary for defining a VideosetDataFrame. 

    Only required argument is list of DLC tracking file names. 

    Any other keyword argument must be either a non-iterable object (e.g. a scalar parameter, like FPS)
    that will be copied and tagged to each of the DLC tracking files, or an iterable object of the same
    length of the list of DLC tracking files. Each element in the iterable will be tagged with the corresponding
    DLC file.

    Args:
        tracking_files: List of DLC tracking .csvs
        **kwargs: described as above

    Returns:
        Dictionary whose keys are DLC tracking file names, and contains a dictionary with key,values containing
        the metadata provided
    """

    metadata = {}
    for fn in tracking_files:
        metadata[fn] = {}
    n_files = len(tracking_files)

    for k,v in kwargs.items():
        if hasattr(v, '__len__'):
            if len(v) != n_files:
                _add_item_to_dict(tracking_files, metadata, k, v)
            else:
                _add_items_to_dict(tracking_files, metadata, k, v)
        else:
            _add_item_to_dict(tracking_files, metadata, k, v)

    return metadata

class VideosetDataFrame(MLDataFrame):
    def __init__(self, metadata : dict, 
                       label_key : dict = None, 
                       part_renamer : dict = None,
                       animal_renamer : dict = None):
        """Houses DLC tracking data and behavior annotations in pandas DataFrame for ML, along with relevant metadata, features and behavior annotation labels.

        Args:
            metadata: Dictionary whose keys are DLC tracking csvs, and value is a dictionary of associated metadata
                for that video. Most easiest to create with 'clone_metadata'. 
                Required keys are: ['scale', 'fps', 'units', 'resolution', 'label_files']
            label_key: Default None. Dictionary whose keys are behavior labels and values are integers 
            part_renamer: Default None. Dictionary that can rename body parts from tracking files if needed (for feature creation, e.g.)
            animal_renamer: Default None. Dictionary that re rename animals from tracking files if needed
        """
        self.req_cols = ['frame_length', 'fps', 'units', 'resolution']

        self.data = pd.DataFrame()
        self.label_key = label_key
        if self.label_key:
            self.reverse_label_key = {v:k for k,v in self.label_key.items()}
        else: 
            self.reverse_label_key = None

        if self._validate_metadata(metadata):   
            self.metadata = metadata
        else:
            raise ValueError("Metadata not properly formatted. See docstring.")
    
        if len(metadata) > 0:
            self._load_tracks(part_renamer, animal_renamer) 
            #By default make these tracks the features... 
            #useless for ML but just to have something to start with
            #This will be updated once some features to do ML with have been computed
            self._load_labels(set_as_label = True)
        else:
            self.raw_track_columns = None

        self.feature_cols = None

    def _setup_default_cv_folds(self):
        default_fold_cols = [f'fold{idx}' for idx in range(self.n_videos)]
        self.fold_cols = default_fold_cols 
        raise NotImplementedError

    def _validate_metadata(self, metadata):
        for fn in metadata:
            checks = [col in metadata[fn].keys() for col in self.req_cols]
            if sum(checks) < len(self.req_cols):
                return False
        return True

    @property
    def videos(self):
        return list(self.metadata.keys())

    @property
    def n_videos(self):
        return len(self.metadata)

    @property
    def group(self):
        return self.data.filename.to_numpy()

    def activate_features_by_name(self, name : str) -> list:
        """Add already present columns in data frame to the feature set. 
        
        Args:
            name: string for pattern matching -- any feature that starts with this string will be added

        Returns:
            List of matched columns (may include columns that were already activated).
        """

        matched_cols = [l for l in self.data.columns if re.match(f"^{name}", l)]
        if self.feature_cols is not None:
            self.feature_cols = uniquifier(list(self.feature_cols) + list(matched_cols))
        else:
            self.feature_cols = matched_cols
        return matched_cols

    def remove_features_by_name(self, name : str) -> list:
        """Remove columns from the feature set. 
        
        Args:
            name: string for pattern matching -- any feature that starts with this string will be removed

        Returns:
            List of removed columns.
        """
        matched_cols = [l for l in self.feature_cols if re.match(f"^{name}", l)]
        removed_cols = self.remove_feature_cols(matched_cols)
        return removed_cols

    #Set features by individual or by group names
    def add_features(self, feature_maker : Features, 
                           featureset_name : str, 
                           add_to_features = False, 
                           **kwargs) -> list:
        """Compute features to dataframe using Feature object. 'featureset_name' will be prepended to new columns, followed by a double underscore. 

        Args:
            featuremaker: A Feature object that houses the feature-making function to be executed and a list of required columns that must in the dataframe for this to work
            featureset_name: Name to prepend to the added features 
            add_to_features: Whether to add to list of active features (i.e. will be returned by the .features property)
        Returns:
            List of new columns that are computed
        """
        new_features = feature_maker.make(self, **kwargs)

        #Prepend these column names w featureset-name__feature-name
        new_feat_cols = list(new_features.columns)
        new_feat_cols = [str(featureset_name) + '__' + str(i) for i in new_feat_cols]
        new_features.columns = new_feat_cols

        self.data = pd.concat([self.data.reset_index(drop = True), 
                               new_features.reset_index(drop = True)], axis = 1)
        if add_to_features:
            if self.feature_cols is not None:
                self.feature_cols = list(self.feature_cols) + list(new_features.columns)
            else:
                self.feature_cols = new_features.columns
        return list(new_features.columns)

    def remove_feature_cols(self, col_names):
        new_col_names = [i for i in self.feature_cols if i not in col_names]
        removed = [i for i in self.feature_cols if i in col_names]
        self.feature_cols = new_col_names
        return removed

    def _load_tracks(self, part_renamer, animal_renamer):
        #For the moment only supports DLC
        return self._load_dlc_tracks(part_renamer, animal_renamer)

    def _load_dlc_tracks(self, part_renamer, animal_renamer):
        """Add DLC tracks to DataFrame"""
        df = pd.DataFrame()
        dfs = []
        col_names_old = None
        #Go through each video file and load DLC tracks
        for fn in self.metadata.keys():
            df_fn, body_parts, animals, col_names, scorer = read_DLC_tracks(fn, 
                                                                            part_renamer, 
                                                                            animal_renamer)
            n_rows = len(df_fn)
            dfs.append(df_fn)
            self.metadata[fn]['duration'] = n_rows/self.metadata[fn]['fps']
            self.metadata[fn]['scorer'] = scorer
            if col_names_old is not None:
                if col_names != col_names_old:
                    raise RuntimeError("DLC files have different columns. Must all be from same project")
            col_names_old = col_names
        df = pd.concat(dfs, axis = 0)
        df = df.reset_index(drop = True)
        self.body_parts = body_parts
        self.animals = animals
        self.animal_setup = {'mouse_ids': animals, 'bodypart_ids': body_parts, 'colnames': col_names}
        self.raw_track_columns = col_names
        self.data = df

    def _load_labels(self, col_name = 'label', set_as_label = False):
        #For the moment only BORIS support
        return self._load_labels_boris(col_name, set_as_label)

    def _load_labels_boris(self, col_name = 'label', set_as_label = False):
        """Add behavior label data to DataFrame"""

        for vid in self.metadata:
            if 'label_files' in self.metadata[vid]:

                fn_in = self.metadata[vid]['label_files']
                fps = self.metadata[vid]['fps']
                duration = self.metadata[vid]['duration']

                ground_truth = read_boris_annotation(fn_in, fps, duration)

                #Add this to data
                self.data.loc[self.data['filename'] == vid, col_name] = ground_truth

        if set_as_label:
            self.label_cols = col_name

    def save(self, fn_out):
        """Save object"""
        with open(fn_out,'wb') as file:
            file.write(pickle.dumps(self.__dict__))

    def load(self, fn_in):
        """Load from file"""
        with open(fn_in, 'rb') as file:
            dataPickle = file.read()
        self.__dict__ = pickle.loads(dataPickle)

    def make_movie(self, label_columns, path_out : str, video_filenames = None) -> None:
        """Given columns indicating behavior predictions or whatever else, make a video
        with these predictions overlaid. 

        VideosetDataFrame must have the keys 'video_file', so that the video associated with each set of DLC tracks is known.

        Args:
            label_columns: list or dict of columns whose values to overlay on top of video. If dict, keys are the columns and values are the print-friendly version.
            path_out: the directory to output the videos too
            video_filenames: list or string. The 

        Returns:
            None. Videos are saved to 'path_out'
        """

        if not checkFFMPEG():
            print("Cannot find ffmpeg in path. Please install.")
            return

        #Text parameters (units in pixels)
        #TODO  
        # Add these to some config.yaml (or json) file
        y_offset = 60
        y_inc = 30
        text_color = 'green'
        font_size = 36

        if type(video_filenames) is str:
            video_filenames = [video_filenames]
        if not video_filenames:
            video_filenames = self.videos
        if type(label_columns) is list:
            label_columns = {k:k for k in label_columns}

        for video in video_filenames:
            rate = 1/self.metadata[video]['fps']
            vid_in = self.metadata[video]['video_files']
            file_out = os.path.splitext(os.path.basename(vid_in))[0] + '_'.join(label_columns.keys()) + '.mp4'
            vid_out = os.path.join(path_out, file_out)
            label_strings = []
            for idx, (col,print_label) in enumerate(label_columns.items()):
                this_y_offset = y_offset + y_inc*idx
                behavior_pairs = _make_dense_values_into_pairs(self.data.loc[self.data.filename == video, col], rate)
                this_label_string = ','.join([f"drawtext=text=\'{print_label}\':x=90:y={this_y_offset}:fontsize={font_size}:fontcolor={text_color}:enable=\'between(t,{pair[0]},{pair[1]})\'" for pair in behavior_pairs])
                label_strings.append(this_label_string)
            label_string = ','.join(label_strings)

            #ffmpeg is the fastest way to add this information to a video
            #Prepare the ffmpeg command
            cmd = f'ffmpeg -y -i {vid_in} -vf "{label_string}" {vid_out}'
            os.system(cmd)
            
def load_videodataset(fn_in : str) -> VideosetDataFrame:
    """Load from file"""
    with open(fn_in, 'rb') as file:
        dataPickle = file.read()
    new_obj = VideosetDataFrame({})
    new_obj.__dict__ = pickle.loads(dataPickle)
    return new_obj


def _make_dense_values_into_pairs(predictions, rate):
    #Put into start/stop pairs
    pairs = []
    in_pair = False
    start = 0
    for idx, behavior in enumerate(predictions):
        if behavior == 0:
            if in_pair:
                pairs.append([start, idx*rate])
            in_pair = False
        if behavior == 1:
            if not in_pair:
                start = idx*rate
            in_pair = True

    if in_pair:
        pairs.append([start, idx*rate])
    return pairs
    


import numpy as np
import json 
import pandas as pd
import tqdm
import os
from copy import deepcopy
import warnings

from behaveml.dl.dl_models import build_baseline_model
from behaveml.dl.dl_generators import MABe_Generator, features_identity
from behaveml.dl.grid_searches import sweeps_baseline, feature_spaces

try:
    import keras
    has_keras = True
except ImportError:
    warnings.warn("Keras not found. Deep learning-based features are not available", RuntimeWarning)
    has_keras = False

THIS_FILE_DIR = os.path.dirname(os.path.realpath(__file__))

def seed_everything(seed = 2012):
    np.random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

seed_everything()

#Implement cross validation, instead of a single validation set
#Product predictions on the held out data, all merged into one frame
class Trainer(object):
    def __init__(self, *,
               feature_dim, 
               num_classes,
               test_data = None,
               class_to_number=None,
               past_frames=0, 
               future_frames=0,
               frame_gap=1, 
               use_conv=False, 
               build_model = build_baseline_model,
               Generator = MABe_Generator,
               use_callbacks = False,
               learning_decay_freq = 10,
               featurizer = features_identity):

        flat_dim = np.prod(feature_dim)
        if use_conv:
            input_dim = ((past_frames + future_frames + 1), flat_dim,)
        else:
            input_dim = (flat_dim * (past_frames + future_frames + 1),)

        self.learning_decay_freq = learning_decay_freq
        self.input_dim = input_dim
        self.use_conv=use_conv
        self.num_classes=num_classes
        self.build_model = build_model 
        self.use_callbacks = use_callbacks

        c2n = {'other': 0,'investigation': 1, 'attack' : 2, 'mount' : 3}
        self.class_to_number = class_to_number or c2n

        if test_data is not None:
            self.test_generator = Generator(test_data, 
                                            batch_size=8192, 
                                            dim=input_dim, 
                                            num_classes=num_classes, 
                                            past_frames=past_frames,
                                            future_frames=future_frames,
                                            use_conv=use_conv,
                                            class_to_number=self.class_to_number,
                                            frame_gap=frame_gap,
                                            augment=False,
                                            shuffle=False,
                                            mode='predict',
                                            featurize = featurizer)

    def delete_model(self):
        self.model = None

    def initialize_model(self, **kwargs):
        self.model = self.build_model(input_dim = self.input_dim, **kwargs)

    def _set_model(self, model):
        """ Set an external, provide initialized and compiled keras model """
        self.model = model

    def train(self, model_params, class_weight=None, n_folds = 5):

        kwargs = {}
        if class_weight is not None:
            if type(class_weight) is dict:
                kwargs['class_weight'] = class_weight

        all_test_pred_probs = {}

        #Repeat this fit process:
        for count in range(n_folds):

            #Reinit model to start training again
            self.initialize_model(**model_params)

            #Instead of training, just load the model
            model_path = os.path.join(THIS_FILE_DIR, 
                                      'pretrained_models', 
                                      f'task1_model_fold_{count}/the_model/variables/variables')
            loaded_model = self.model
            loaded_model.load_weights(model_path).expect_partial()
            self.model = loaded_model 

            #Once trained we can do the inference on the test data:
            test_pred_probs = self.get_test_prediction_probabilities()

            #Add this to the all_test_pred_prob
            for k in test_pred_probs:
                if k in all_test_pred_probs:
                    all_test_pred_probs[k] += test_pred_probs[k]
                else:
                    all_test_pred_probs[k] = test_pred_probs[k]

        #Once we've done that
        #Take the average over all these predictions:
        for k in all_test_pred_probs:
            all_test_pred_probs[k] /= n_folds

        return all_test_pred_probs

    def get_test_prediction_probabilities(self):
        all_test_preds = {}
        for vkey in self.test_generator.video_keys:
            nframes = self.test_generator.seq_lengths[vkey]
            all_test_preds[vkey] = np.zeros((nframes,4), dtype=np.float32)

        for X, vkey_fi_list in tqdm.tqdm(self.test_generator):
            test_pred = self.model.predict(X)
            #test_pred = np.argmax(test_pred, axis=-1)

            if len(test_pred.shape) > 2:
                test_pred = self.get_final_val_probabilities(test_pred)

            for p, (vkey, fi) in zip(test_pred, vkey_fi_list):
                all_test_preds[vkey][fi] = p
        return all_test_preds

def normalize_data(orig_pose_dictionary):
    for key in orig_pose_dictionary:
        X = orig_pose_dictionary[key]['keypoints']
        X = X.transpose((0,1,3,2)) #last axis is x, y coordinates
        #Original MARS normalization
        X[..., 0] = X[..., 0]/1024
        X[..., 1] = X[..., 1]/570
        orig_pose_dictionary[key]['keypoints'] = X
    return orig_pose_dictionary

def run_task(vocabulary, test_data, config_name,
              build_model, skip_test_prediction=False, seed=2021,
              Generator = MABe_Generator, use_callbacks = False, params = None, 
              use_conv = True):

    if params is None:
        if config_name is None:
            raise ValueError("Provide one of params dictionary or config_name with path to json file")
        with open(config_name, 'r') as fp:
            params = json.load(fp)

    normalize = params["normalize"]
    params["seed"] = seed
    seed_everything(seed)

    if "steps_per_epoch" in params:
        steps_per_epoch = params["steps_per_epoch"]
        if steps_per_epoch == -1:
            steps_per_epoch = None
    else:
        steps_per_epoch = None

    features = params['features']
    feature_dim = feature_spaces[features][1]
    featurizer = feature_spaces[features][0]

    if normalize:
        if not skip_test_prediction:
            test_data = normalize_data(deepcopy(test_data))
        else:
            test_data = None

    num_classes = 4

    class_to_number = vocabulary

    trainer = Trainer(feature_dim=feature_dim, 
                      num_classes=num_classes,
                      test_data = test_data,
                      class_to_number=class_to_number,
                      past_frames=params['past_frames'], 
                      future_frames=params['future_frames'],
                      frame_gap=params['frame_gap'],
                      use_conv=use_conv,
                      build_model = build_model,
                      Generator = Generator,
                      use_callbacks = use_callbacks,
                      learning_decay_freq=params['learning_decay_freq'],
                      featurizer = featurizer)

    #Extract model param dictionary
    model_params = {}
    for k in params:
        if 'model_param' in k:
            k_ = k.split('__')[1]
            model_params[k_] = params[k]

    if params['reweight_loss'] is True:
        class_weight = 1/trainer.train_generator.class_weights
    else:
        class_weight = None

    model_params['class_weight'] = class_weight
    all_test_probs = trainer.train(class_weight = class_weight, model_params = model_params)
    return all_test_probs

def lrs(epoch, lr, freq = 10):
    if (epoch % freq) == 0 and epoch > 0:
        lr /= 3 
    return lr

def convert_to_mars_format(df, colnames, animal_setup):
    n_animals = len(animal_setup['mouse_ids'])
    n_body_parts = len(animal_setup['bodypart_ids'])
    pose_dict = {}
    videos = np.unique(df.filename)
    for vid in videos:
        pose_dict[vid] = {'annotator_id': 0}
        keypoints = df.loc[df.filename == vid, colnames]
        n_rows = keypoints.shape[0]
        pose_dict[vid]['keypoints'] = keypoints.to_numpy().reshape((n_rows, n_animals, 2, n_body_parts))
    return pose_dict

#Basically, undo the change above
def convert_to_pandas_df(data, colnames = None):
    dfs = []
    for vid in data:
        df = pd.DataFrame(data[vid], columns = colnames)
        dfs.append(df)
    final_df = pd.concat(dfs, axis = 0)
    return final_df

def compute_dl_probability_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs):

    if not has_keras:
        raise RuntimeError("Keras not found. Deep learning-based features are not available")

    test_data = convert_to_mars_format(df, raw_col_names, animal_setup)
    parametersweep = 'test_run_distances'
    config_name = os.path.join(THIS_FILE_DIR, 'dl_baseline_settings.json')
    build_model = build_baseline_model
    Generator = MABe_Generator
    use_callbacks = False
    sweeps = sweeps_baseline

    #Load default config
    with open(config_name, 'r') as fp:
        config = json.load(fp)
    if 'model_param__layer_channels' in config:
        config['model_param__layer_channels'] = tuple(config['model_param__layer_channels'])

    #Modify to setup parameter sweep    
    for k in sweeps[parametersweep][1]:
        config[k] = sweeps[parametersweep][1][k]

    arguments = {'config_name': config_name,
                 'build_model': build_model,
                 'Generator': Generator,
                 'use_callbacks': use_callbacks}

    config['future_frames'] = 50
    config['past_frames'] = 50
    config['model_param__learning_rate'] = 0.0001

    vocab = {'attack': 0, 'investigation': 1, 'mount': 2, 'other': 3}
                            
    all_test_probs = run_task(vocabulary=vocab,
                                test_data=test_data,
                                config_name = None,
                                build_model = arguments['build_model'],
                                seed=2021,
                                Generator = arguments['Generator'],
                                use_callbacks = arguments['use_callbacks'],
                                params = config)

    colnames = [f'prob_{k}' for k in vocab.keys()]
    test_df = convert_to_pandas_df(all_test_probs, colnames = colnames)
    return test_df


""" Functions to take pose tracks and compute a set of features from them """

from typing import Callable

from behaveml.dl.dl_features import compute_dl_probability_features
from behaveml.mars_features import compute_mars_features

default_tracking_columns = ['adult_x_nose', 'adult_x_leftear', 'adult_x_rightear', 'adult_x_neck',
                            'adult_x_lefthip', 'adult_x_righthip', 'adult_x_tail', 'adult_y_nose',
                            'adult_y_leftear', 'adult_y_rightear', 'adult_y_neck',
                            'adult_y_lefthip', 'adult_y_righthip', 'adult_y_tail',
                            'juvenile_x_nose', 'juvenile_x_leftear', 'juvenile_x_rightear',
                            'juvenile_x_neck', 'juvenile_x_lefthip', 'juvenile_x_righthip',
                            'juvenile_x_tail', 'juvenile_y_nose', 'juvenile_y_leftear',
                            'juvenile_y_rightear', 'juvenile_y_neck', 'juvenile_y_lefthip',
                            'juvenile_y_righthip', 'juvenile_y_tail']

class Features(object):
    def __init__(self, feature_maker : Callable, required_columns : list, **kwargs):
        self.required_columns = required_columns
        self.feature_maker = feature_maker

    def make(self, vdf, **kwargs):
        #Validate columns:
        checks = [col in self.required_columns for col in vdf.data.columns]
        if sum(checks) < len(self.required_columns):
            raise RuntimeError("VideoDataFrame doesn't have necessary columns to compute this set of features.")
        new_features = self.feature_maker(vdf.data, self.required_columns, vdf.animal_setup, **kwargs)
        return new_features

mars_feature_maker = Features(compute_mars_features, default_tracking_columns)
cnn_probability_feature_maker = Features(compute_dl_probability_features, default_tracking_columns)


import pandas as pd 
import numpy as np
from behaveml.video import VideosetDataFrame

def interpolate_lowconf_points(vdf : VideosetDataFrame,
                               filter_out_lowconf : bool = True,
                               filter_out_toofast : bool = True,
                               conf_threshold : float = 0.9,
                               jump_dur : int = 5,
                               speed_threshold : float = 5,
                               in_place = True) -> pd.DataFrame:
    """Interpolate raw tracking points if their probabilities are available.

    Args:
        vdf: VideosetDataFrame containing the tracks to interpolate
        filter_out_lowconf: default True. Whether to filter out low confidence points
        filter_out_toofast: default True. Whether to filter out tracks that jump too far in a number of frames
        conf_threshold: default 0.9. Confidence below which to count as uncertain, and to interpolate its value instead
        jump_dur: default 5. Number of frames to compute velocity which is used as basis for filtering out jumps
        speed_threshold: default 5. Number of pixels to 
        in_place: default True. Whether to replace data in place

    Returns:
        Pandas dataframe with the filtered raw columns. Returns None if opted for in_place modification
    """

    df_filtered = []

    for fn_in in vdf.videos:
        print("processing", fn_in)

        if not in_place:
            df_filter_low_conf = vdf.data.loc[vdf.data.filename == fn_in].copy()
        else:
            df_filter_low_conf = vdf.data
        
        if filter_out_lowconf:
            for m in vdf.animals:
                for bp in vdf.body_parts:
                    low_conf = df_filter_low_conf.loc[vdf.data.filename == fn_in, '_'.join(['likelihood', m, bp])] < conf_threshold
                    df_filter_low_conf.loc[(vdf.data.filename == fn_in) & low_conf,'_'.join([m, 'x', bp])] = np.nan
                    df_filter_low_conf.loc[(vdf.data.filename == fn_in) & low_conf,'_'.join([m, 'y', bp])] = np.nan
                    
            df_filter_low_conf.loc[(vdf.data.filename == fn_in)] = \
                df_filter_low_conf.loc[(vdf.data.filename == fn_in)].\
                interpolate(axis = 0, method = 'linear', limit_direction = 'both')            

            df_filter_low_conf.loc[(vdf.data.filename == fn_in), vdf.raw_track_columns] = \
                df_filter_low_conf.loc[(vdf.data.filename == fn_in), vdf.raw_track_columns].rolling(window = 3, min_periods = 1).mean()
            
        #TODO
        # Get this working...
        # if filter_out_toofast:
        #     for m in vdf.animals:
        #         for bp in vdf.body_parts:
        #             too_quick = (df_filter_low_conf.loc[:,'_'.join([m, 'x', bp])].diff(jump_dur).abs() > speed_threshold) | \
        #                         (df_filter_low_conf.loc[:,'_'.join([m, 'y', bp])].diff(jump_dur).abs() > speed_threshold) | \
        #                         (df_filter_low_conf.loc[:,'_'.join([m, 'x', bp])].diff(-jump_dur).abs() > speed_threshold) | \
        #                         (df_filter_low_conf.loc[:,'_'.join([m, 'y', bp])].diff(-jump_dur).abs() > speed_threshold)

        #             df_filter_low_conf.loc[too_quick,'_'.join([m, 'x', bp])] = np.nan
        #             df_filter_low_conf.loc[too_quick,'_'.join([m, 'y', bp])] = np.nan

        #     #And now filter again
        #     df_filter_low_conf = df_filter_low_conf.interpolate(axis = 0, method = 'polynomial', order = 1)
        #     df_filter_low_conf = df_filter_low_conf.rolling(window = 3, min_periods = 1).mean()

        df_filter_low_conf = df_filter_low_conf[vdf.raw_track_columns + ['filename', 'frame']]
        df_filtered.append(df_filter_low_conf)

    df_filtered = pd.concat(df_filtered, axis = 0)

    if not in_place:
        return df_filtered
    else:
        return None

#Try using the fancyimpute package....
# #df_filter_low_conf = pd.DataFrame(NuclearNormMinimization().fit_transform(df_filter_low_conf.to_numpy()), 
# #                                 index = df_filter_low_conf.index, 
# #                                 columns = df_filter_low_conf.columns)

# df_filter_low_conf = pd.DataFrame(KNN(k=7).fit_transform(df_filter_low_conf.to_numpy()), 
#                                 index = df_filter_low_conf.index, 
#                                 columns = df_filter_low_conf.columns)

# #X_incomplete_normalized = BiScaler().fit_transform(df_filter_low_conf.to_numpy())
# #df_filter_low_conf = pd.DataFrame(SoftImpute().fit_transform(X_incomplete_normalized),
# #                                 index = df_filter_low_conf.index,
# #                                 columns = df_filter_low_conf.columns)



import numpy as np 
from collections import Counter 
import pandas as pd 
import warnings

try:
    import keras
    has_keras = True
except ImportError:
    warnings.warn("Keras not found. Deep learning-based features are not available", RuntimeWarning)
    has_keras = False

#from lib.helper import colnames 
from .feature_engineering import make_features_mars, make_features_mars_distr

def make_df(pts, colnames = None):
    df = []
    for idx in range(len(pts)):
        data = pts[idx].flatten()
        df.append(list(data))
    if colnames:
        return pd.DataFrame(df, columns = colnames)
    else:
        return pd.DataFrame(df)

def features_identity(inputs):
    return inputs, inputs.shape[1:]

def features_via_sklearn(inputs, featurizer):
    #Use the ML functions to turn this into a pandas data table
    df = make_df(inputs)
    features_df, _, _ = featurizer(df)
    features = np.array(features_df)
    return features, features.shape

features_mars = lambda x: features_via_sklearn(x, make_features_mars)

#features_mars_no_shift = lambda x: features_via_sklearn(x, make_features_mars_no_shift)

features_mars_distr = lambda x: features_via_sklearn(x, make_features_mars_distr)

def features_distances(inputs):

    #inputs.shape (4509, 2,7,2) = (frame, mouse ID, body part, x/y)

    features = []    
    ##Make the distance features
    for i in range(7):
        for j in range(7):
            if i < j:
                for mouse_id in range(2):
                    #We can compute the intra-mouse difference
                    f1x = inputs[:,mouse_id,i,0]
                    f2x = inputs[:,mouse_id,j,0]
                    f1y = inputs[:,mouse_id,i,1]
                    f2y = inputs[:,mouse_id,j,1]
                    features.append(np.sqrt((f1x - f2x)**2 + (f1y - f2y)**2))
            #Inter-mouse difference
            f1x = inputs[:,0,i,0]
            f2x = inputs[:,1,j,0]
            f1y = inputs[:,0,i,1]
            f2y = inputs[:,1,j,1]
            features.append(np.sqrt((f1x - f2x)**2 + (f1y - f2y)**2))

    features = np.vstack(features).T

    return features, features.shape[1:]

def features_distances_normalized(inputs):

    #inputs.shape (4509, 2,7,2) = (frame, mouse ID, body part, x/y)

    features = []    
    for i in range(7):
        for j in range(7):
            if i < j:
                for mouse_id in range(2):
                    #Compute the intra-mouse difference
                    f1x = inputs[:,mouse_id,i,0]
                    f2x = inputs[:,mouse_id,j,0]
                    f1y = inputs[:,mouse_id,i,1]
                    f2y = inputs[:,mouse_id,j,1]
                    distances = np.sqrt((f1x - f2x)**2 + (f1y - f2y)**2)
                    distances /= np.linalg.norm(distances)
                    features.append(distances)
            #Inter-mouse difference
            f1x = inputs[:,0,i,0]
            f2x = inputs[:,1,j,0]
            f1y = inputs[:,0,i,1]
            f2y = inputs[:,1,j,1]
            distances = np.sqrt((f1x - f2x)**2 + (f1y - f2y)**2)
            features.append(distances)

    features = np.vstack(features).T

    return features, features.shape[1:]
    #output (4509, X) = (frame, feature) 

if has_keras:
    class MABe_Generator(keras.utils.Sequence):
        def __init__(self, pose_dict, 
                    batch_size, dim, 
                    use_conv, num_classes, augment=False,
                    class_to_number=None,
                    past_frames=0, future_frames=0, 
                    frame_gap=1, shuffle=False,
                    mode='fit', featurize = features_identity):
            self.batch_size = batch_size
            self.featurize = featurize
            self.video_keys = list(pose_dict.keys())
            self.dim = dim
            self.use_conv = use_conv
            self.past_frames = past_frames
            self.future_frames = future_frames
            self.frame_gap = frame_gap
            self.shuffle = shuffle
            self.num_classes=num_classes
            self.augment = augment
            self.mode = mode

            self.class_to_number = class_to_number

            self.video_indexes = []
            self.frame_indexes = []
            self.X = {}
            if self.mode == 'fit':
                self.y = []
            self.pad = self.past_frames * self.frame_gap
            future_pad = self.future_frames * self.frame_gap
            pad_width = (self.pad, future_pad), (0, 0), (0, 0), (0, 0)
            self.seq_lengths = {}
            for vc, key in enumerate(self.video_keys):
                if self.mode == 'fit':
                    anno = pose_dict[key]['annotations']
                    self.y.extend(anno)
                nframes = len(pose_dict[key]['keypoints'])
                self.video_indexes.extend([vc for _ in range(nframes)])
                self.frame_indexes.extend(range(nframes))
                self.X[key],_ = self.featurize(np.pad(pose_dict[key]['keypoints'], pad_width))
                self.seq_lengths[key] = nframes
            
            if self.mode == 'fit':
                self.y = np.array(self.y)
                #Compute class weights:
                class_weights = np.zeros(self.num_classes)
                counts = Counter(self.y)
                for c in range(num_classes):
                    class_weights[c] = counts[c]
                class_weights /= np.sum(class_weights)
                self.class_weights = class_weights
            else:
                self.class_weights = np.ones(num_classes)
            
            self.X_dtype = self.X[key].dtype

            self.indexes = list(range(len(self.frame_indexes)))

            if self.mode == 'predict':
                extra_predicts = -len(self.indexes) % self.batch_size # So that last part is not missed
                self.indexes.extend(self.indexes[:extra_predicts])
                self.indexes = np.array(self.indexes)
            
            self.on_epoch_end()

        def __len__(self):
            return len(self.indexes) // self.batch_size

        def augment_fn(self, x):
            # Rotate
            angle = (np.random.rand()-0.5) * (np.pi * 2)
            c, s = np.cos(angle), np.sin(angle)
            rot = np.array([[c, -s], [s, c]])
            x = np.dot(x, rot)

            # Shift - All get shifted together
            shift = (np.random.rand(2)-0.5) * 2 * 0.25
            x = x + shift
            return x

        def __getitem__(self, index):
            bs = self.batch_size
            indexes = self.indexes[index*bs:(index+1)*bs]
            X = np.empty((bs, *self.dim), self.X_dtype)
            if self.mode == 'predict':
                vkey_fi_list = []
            for bi, idx in enumerate(indexes):
                vkey = self.video_keys[self.video_indexes[idx]]
                fi = self.frame_indexes[idx]
                if self.mode == 'predict':
                    vkey_fi_list.append((vkey, fi))
                fi = fi + self.pad
                start = fi - self.past_frames*self.frame_gap
                stop = fi + (self.future_frames + 1)*self.frame_gap
                assert start >= 0

                Xi = self.X[vkey][start:stop:self.frame_gap].copy()
            
                if self.augment:
                    Xi = self.augment_fn(Xi)
                X[bi] = np.reshape(Xi, self.dim)

            if self.mode == 'fit':
                y_vals = self.y[indexes]
                # Converting to one hot because F1 callback needs one hot
                y = np.zeros( (bs,self.num_classes), np.float32)
                y[np.arange(bs), y_vals] = 1
                return X, y

            elif self.mode == 'predict':
                return X, vkey_fi_list

        def on_epoch_end(self):
            if self.shuffle == True:
                np.random.shuffle(self.indexes)
else:
    class MABe_Generator(object):
        pass


def _exec_php(cmd):
    from subprocess import Popen, PIPE, STDOUT
    p = Popen(cmd, shell=False, stdout=PIPE, stderr=STDOUT)
    return [p.wait(), p.stdout.readlines()]
    
def checkFFMPEG() -> bool:
    """
    Check for ffmpeg dependencies

    Returns:
        True if can find ffmpeg in path, false otherwise
    """
    try:
        return_value = _exec_php(['ffmpeg', '-version', '2>&1'])[0]
        if return_value == 0: return True
    except Exception:
        pass

    return False 